@@ -1,20 +1,19 @@
-import argparse
 import copy
 import json
 import os
 import subprocess
 import sys
-import textwrap
-import types
+import uuid
+from abc import ABC, abstractmethod
 from datetime import datetime
 from json import JSONEncoder
-from unittest.mock import patch
 from pathlib import Path
 from pprint import pformat
-from typing import Any, Dict, List, Optional, Set, Union, Sequence, DefaultDict, Tuple
+from typing import Any, Dict, List, Optional, Sequence, Tuple, Union
+from unittest.mock import patch
 
-
-__version__ = "0.2.0"
+__version__ = "2.0.2"
+DATETIME_STRING_FORMAT = "%Y-%m-%dT%H-%M-%S"
 
 
 class Note(dict):
@@ -30,6 +29,7 @@ class Note(dict):
     _start_datetime_key = "start_datetime"
     _end_datetime_key = "end_datetime"
 
+    _identifier_key = "identifier"
     _git_key = "git"
     _python_path_key = "python_path"
 
@@ -41,8 +41,8 @@ class Note(dict):
         such as time and date, last commit and current branch of git repository,
         path to Python executable, etc.
 
-        The most commonly used dictionary keys can be accessed as attributes
-        for better auto-completion support and readability.
+        All initial dictionary keys can be accessed as attributes
+        for better auto-completion support and code readability.
 
 
         Parameters
@@ -52,27 +52,34 @@ class Note(dict):
         content : Optional[Dict[str, dict]], optional (default=None)
             Mainly for internal use to create Note instances out of loaded
             dictionaries from a Store. If content is passed, no additional information
-            is added on instantiation of the class (e.g. no start datetime).
+            is added on instantiation of the class (e.g. no start datetime,
+            identifier, ...), meaning that these attributes already need to be
+            present in the passed in dictionary.
         """
         if content is not None:
             super().__init__(content)
             self._content_passed = True
         else:
             self._content_passed = False
-            self[self._text_key] = text
+            self.text = text
             self._set_up_initial_structure()
             self._start()
 
+    def _set_identifier(self) -> None:
+        self.identifier = str(uuid.uuid4())
+
     def _set_up_initial_structure(self) -> None:
-        self[self._model_key] = None
-        self[self._parameters_key] = {}
-        self[self._features_key] = self._initial_features_structure()
-        self[self._target_key] = None
-        self[self._metrics_key] = {}
-        self[self._info_key] = {}
-        self[self._start_datetime_key] = None
-        self[self._end_datetime_key] = None
-        self[self._python_path_key] = self._python_executable_path()
+        self.model = None
+        self.parameters = {}  # type: dict
+        self.features = self._initial_features_structure()
+        self.target = None
+        self.metrics = {}  # type: dict
+        self.info = {}  # type: dict
+        self.start_datetime = None
+        self.end_datetime = None
+        self._set_identifier()
+        self.python_path = self._python_executable_path()
+        self.git = {}  # type: dict
 
     def _initial_features_structure(self) -> dict:
         """This method can easily be overwritten to return a different
@@ -85,34 +92,34 @@ class Note(dict):
         return sys.executable
 
     def _start(self) -> None:
-        self[self._start_datetime_key] = datetime.now()
+        self.start_datetime = self._current_datetime()
         self._add_git_info()
 
     def end(self) -> None:
         """Adds the current datetime as 'end_datetime' to the note"""
-        self[self._end_datetime_key] = datetime.now()
+        self.end_datetime = self._current_datetime()
+
+    def _current_datetime(self) -> datetime:
+        return datetime.now().replace(microsecond=0)
 
     def _add_git_info(self) -> None:
         if self._is_in_git_repo():
-            git_info = {}
-            git_info["repo_name"] = (
+            self.git["repo_name"] = (
                 subprocess.check_output(["git", "rev-parse", "--git-dir"])
                 .strip()
                 .decode("utf-8")
             )
-            git_info["branch"] = (
+            self.git["branch"] = (
                 subprocess.check_output(["git", "rev-parse", "--abbrev-ref", "HEAD"])
                 .strip()
                 .decode("utf-8")
             )
-            git_info["commit"] = (
+            self.git["commit"] = (
                 subprocess.check_output(["git", "rev-parse", "--short", "HEAD"])
                 .strip()
                 .decode("utf-8")
             )
 
-            self[self._git_key] = git_info
-
     def _is_in_git_repo(self) -> bool:
         """Function based on following stackoverflow answer by tdelaney:
         https://stackoverflow.com/a/19688210
@@ -129,65 +136,129 @@ class Note(dict):
         else:
             return True
 
+    @classmethod
+    def from_note(cls, note: "Note") -> "Note":
+        """Creates a new note from an existing one, taking over its content
+        but setting a new start datetime and identifier.
+
+        Can for example be used when evaluating multiple model parameters
+        (e.g. in a gridsearch setup). Simply create a new note from the original one
+        for each parameter set which should be evaluated and add the parameters
+        and metrics to the new note.
+
+        This method creates a deepcopy of the original one, meaning that modifying
+        the new note should not affect the old one.
+
+        Parameters
+        ----------
+        note : Note
+            Existing ntoe from which the content should be taken over
+
+        Returns
+        -------
+        Note
+        """
+        assert isinstance(note, cls)
+        new_note = copy.deepcopy(note)
+        new_note.start_datetime = new_note._current_datetime()
+        new_note._set_identifier()
+        return new_note
+
     @property
     def identifier(self) -> str:
-        return _format_datetime(self[self._start_datetime_key])
+        return self[self._identifier_key]
+
+    @identifier.setter
+    def identifier(self, value: str):
+        self[self._identifier_key] = value
 
     @property
     def text(self) -> str:
         return self[self._text_key]
 
     @text.setter
-    def text(self, value):
+    def text(self, value: str):
         self[self._text_key] = value
 
     @property
-    def model(self):
+    def model(self) -> str:
         return self[self._model_key]
 
     @model.setter
-    def model(self, value):
+    def model(self, value: str):
         self[self._model_key] = value
 
     @property
-    def metrics(self):
+    def metrics(self) -> dict:
         return self[self._metrics_key]
 
     @metrics.setter
-    def metrics(self, value):
+    def metrics(self, value: dict):
         self[self._metrics_key] = value
 
     @property
-    def parameters(self):
+    def parameters(self) -> dict:
         return self[self._parameters_key]
 
     @parameters.setter
-    def parameters(self, value):
+    def parameters(self, value: dict):
         self[self._parameters_key] = value
 
     @property
-    def features(self):
+    def features(self) -> dict:
         return self[self._features_key]
 
     @features.setter
-    def features(self, value):
+    def features(self, value: dict):
         self[self._features_key] = value
 
     @property
-    def target(self):
+    def target(self) -> str:
         return self[self._target_key]
 
     @target.setter
-    def target(self, value):
+    def target(self, value: str):
         self[self._target_key] = value
 
     @property
-    def info(self):
+    def info(self) -> dict:
         return self[self._info_key]
 
     @info.setter
-    def info(self, value):
-        self[self._info_key] = info
+    def info(self, value: dict):
+        self[self._info_key] = value
+
+    @property
+    def start_datetime(self) -> datetime:
+        return self[self._start_datetime_key]
+
+    @start_datetime.setter
+    def start_datetime(self, value: datetime):
+        self[self._start_datetime_key] = value
+
+    @property
+    def end_datetime(self) -> datetime:
+        return self[self._end_datetime_key]
+
+    @end_datetime.setter
+    def end_datetime(self, value: datetime):
+        self[self._end_datetime_key] = value
+
+    @property
+    def python_path(self) -> str:
+        return self[self._python_path_key]
+
+    @python_path.setter
+    def python_path(self, value: str):
+        self[self._python_path_key] = value
+
+    @property
+    def git(self) -> dict:
+        return self[self._git_key]
+
+    @git.setter
+    def git(self, value: dict):
+        self[self._git_key] = value
 
     def __repr__(self) -> str:
         # Code and idea for patching sorted to prevent sorting by
@@ -198,15 +269,16 @@ class Note(dict):
         return r
 
 
-class BaseStore:
-    """The base store class. This class cannot be used directly and mostly acts
-    as a placeholder which defines the store interface. Inherit from this class if you
+class BaseStore(ABC):
+    """The base store class. This class cannot be used directly and acts
+    as a template which defines the store interface. Inherit from this class if you
     want to implement your own store class.
     """
 
     def __init__(self):
         pass
 
+    @abstractmethod
     def load(self, return_dataframe: bool = False):
         """Should return the full store as List[Note] with the most recent
         note first, or, if return_dataframe=True,
@@ -215,100 +287,125 @@ class BaseStore:
         This method is intended to be implemented by subclasses and so
         raises a NotImplementedError.
         """
-        raise NotImplementedError("Should be implemented by BaseStore subclasses")
-        loaded_notes = self._load()
-        if return_dataframe:
-            loaded_notes = self._to_pandas(loaded_notes)
-        return loaded_notes
+        pass
 
+    @abstractmethod
     def add(self, note: Note):
         """Should add an note to the persistent store implemented by the subclass
 
         This method is intended to be implemented by subclasses and so
         raises a NotImplementedError.
         """
-        raise NotImplementedError("Should be implemented by BaseStore subclasses")
-        note = self._prepare_note_for_storing(note)
+        pass
 
-    def _prepare_note_for_storing(self, note: Note) -> Note:
-        if note[note._end_datetime_key] is None:
-            note.end()
-        return copy.deepcopy(note)
 
-    def _to_pandas(self, notes: List[Note]):
-        try:
-            import pandas as pd
-        except ImportError:
-            raise ImportError(
-                "Pandas is not installed. You can install it via:\n"
-                "conda install pandas\n"
-                "or: pip install pandas"
-            )
-        return pd.DataFrame(copy.deepcopy(self._pandas_dict(notes)))
+def _prepare_note_for_storing(note: Note) -> Note:
+    if note.end_datetime is None:
+        note.end()
+    return copy.deepcopy(note)
 
-    def _pandas_dict(self, notes: List[Note]) -> dict:
-        flat_dicts = []
-        for note in notes:
-            flat_dicts.append(self._flatten_dict(dict(note)))
 
-        column_dict = {}  # type: Dict[str, list]
-        all_column_names = set([col for d in flat_dicts for col, _ in d.items()])
-        for column in all_column_names:
-            column_dict[column] = []
-
-        for d in flat_dicts:
-            for column in all_column_names:
-                column_dict[column].append(d.get(column))
-
-        key_order = (
-            [
-                Note._start_datetime_key,
-                Note._end_datetime_key,
-                Note._text_key,
-                Note._model_key,
-            ]
-            + self._filter_sequence_if_startswith(
-                column_dict.keys(), startswith=Note._metrics_key
-            )
-            + self._filter_sequence_if_startswith(
-                column_dict.keys(), startswith=Note._parameters_key
-            )
-            + self._filter_sequence_if_startswith(
-                column_dict.keys(), startswith=Note._features_key
-            )
-            + [Note._target_key]
-            + self._filter_sequence_if_startswith(
-                column_dict.keys(), startswith=Note._git_key
-            )
+def _to_pandas(notes: List[Note]):
+    try:
+        import pandas as pd  # type: ignore
+    except ImportError:
+        raise ImportError(
+            "Pandas is not installed. You can install it via:\n"
+            "conda install pandas\n"
+            "or: pip install pandas"
         )
-        key_order.extend(key for key in column_dict.keys() if key not in key_order)
-        ordered_dict = {k: column_dict[k] for k in key_order}
-        return ordered_dict
+    return pd.DataFrame(copy.deepcopy(_pandas_dict(notes)))
 
-    def _flatten_dict(self, d: dict, parent_key: str = "", sep: str = ".") -> dict:
-        """Flattens a dictionary by concatenating key names
 
-        Taken from https://stackoverflow.com/a/6027615
-        """
-        items = []  # type: List[tuple]
-        for k, v in d.items():
-            new_key = parent_key + sep + k if parent_key else k
-            if isinstance(v, dict):
-                items.extend(self._flatten_dict(v, new_key, sep=sep).items())
-            else:
-                items.append((new_key, v))
-        return dict(items)
+def _pandas_dict(notes: List[Note]) -> dict:
+    flat_dicts = _flatten_notes(notes)
+    all_keys = _all_keys_from_dicts(flat_dicts)
 
-    def _filter_sequence_if_startswith(
-        self, seq: Sequence[str], startswith: str
-    ) -> List[str]:
-        return [x for x in seq if x.startswith(startswith)]
+    # Create basic structure
+    column_dict = {}  # type: Dict[str, list]
+    for column in all_keys:
+        column_dict[column] = []
+
+    # Fill structure with values
+    for d in flat_dicts:
+        for column in all_keys:
+            column_dict[column].append(d.get(column))
+
+    key_order = _key_order(all_keys)
+    ordered_dict = {k: column_dict[k] for k in key_order}
+    return ordered_dict
+
+
+def _flatten_notes(notes: Sequence[Note]) -> List[Dict[str, Any]]:
+    flat_dicts = []
+    for note in notes:
+        flat_dicts.append(_flatten_dict(dict(note)))
+    return flat_dicts
+
+
+def _all_keys_from_dicts(ds: Sequence[Dict[str, Any]]) -> List[str]:
+    return list(set([col for d in ds for col, _ in d.items()]))
+
+
+def _key_order(
+    keys: Sequence[str], additional_keys_subset: Optional[Sequence[str]] = None
+) -> List[str]:
+    """start_datetime, end_datetime, text, model, and identifier are always first.
+    Afterwards, either all keys are added in order of metrics, parameters,
+    features, git, and others, or only the passed in categories
+    from additional_keys_subset. additional_keys_subset can hereby just be
+    the start of the strings, e.g. ["metrics", "parameters]"
+    """
+    key_order = [
+        Note._start_datetime_key,
+        Note._end_datetime_key,
+        Note._text_key,
+        Note._model_key,
+        Note._identifier_key,
+    ]
+    if additional_keys_subset is None:
+        key_order += (
+            _filter_sequence_if_startswith(keys, startswith=Note._metrics_key)
+            + _filter_sequence_if_startswith(keys, startswith=Note._parameters_key)
+            + _filter_sequence_if_startswith(keys, startswith=Note._features_key)
+            + [Note._target_key]
+            + _filter_sequence_if_startswith(keys, startswith=Note._info_key)
+            + _filter_sequence_if_startswith(keys, startswith=Note._git_key)
+            + [Note._python_path_key]
+        )
+        key_order.extend(sorted([k for k in keys if k not in key_order]))
+    else:
+        for k in additional_keys_subset:
+            key_order += _filter_sequence_if_startswith(keys, startswith=k)
+    return key_order
+
+
+def _flatten_dict(d: dict, parent_key: str = "", sep: str = ".") -> dict:
+    """Flattens a dictionary by concatenating key names
+
+    Taken from https://stackoverflow.com/a/6027615
+    """
+    items = []  # type: List[Tuple[Any, Any]]
+    for k, v in d.items():
+        new_key = parent_key + sep + k if parent_key else k
+        if isinstance(v, dict):
+            items.extend(_flatten_dict(v, new_key, sep=sep).items())
+        else:
+            items.append((new_key, v))
+    return dict(items)
+
+
+def _filter_sequence_if_startswith(seq: Sequence[str], startswith: str) -> List[str]:
+    return sorted([x for x in seq if x.startswith(startswith)])
 
 
 class Store(BaseStore):
-    def __init__(self, path: Union[str, Path]) -> None:
-        """Stores and loads Note instances in and from a json file
+    """Main purpose is to store Note instances in a json file. Additional methods are
+    provided to load all notes, as well as update or remove specified notes.
+    """
 
+    def __init__(self, path: Union[str, Path]) -> None:
+        """
         Parameters
         ----------
         path : Union[str, Path]
@@ -343,7 +440,7 @@ class Store(BaseStore):
         """
         loaded_notes = self._load()
         if return_dataframe:
-            loaded_notes = self._to_pandas(loaded_notes)
+            loaded_notes = _to_pandas(loaded_notes)
         return loaded_notes
 
     def _load(self) -> List[Note]:
@@ -352,7 +449,10 @@ class Store(BaseStore):
         return self._sort_notes(notes)
 
     def add(self, note: Note) -> None:
-        """Adds the given note to the .json file of the store
+        """Adds the given note to the .json file of the store.
+
+        Before storing the note, the .end method of it is called, if
+        not already done previously.
 
         Parameters
         ----------
@@ -371,10 +471,11 @@ class Store(BaseStore):
         all_notes = self.load()
         if self._notes_are_subset(notes_subset=[note], all_notes=all_notes):
             raise Exception(
-                f"The identifier for the note '{note.identifier}' already exists in the store."
+                f"The identifier for the note '{note.identifier}' "
+                + "already exists in the store."
                 + " The note was not added."
             )
-        note = self._prepare_note_for_storing(note)
+        note = _prepare_note_for_storing(note)
         all_notes.append(note)
         self._save_notes(all_notes)
 
@@ -478,19 +579,24 @@ class Store(BaseStore):
         self._json_dump(raw_dicts, self.path)
 
     def _sort_notes(self, notes: List[Note]) -> List[Note]:
-        """Sorted by start datetime. Most recent note first"""
-        return list(sorted(notes, key=lambda x: x[x._start_datetime_key], reverse=True))
+        """Sorted by end datetime (descending order, i.e. newest first)
+        and if there is a tie also by the identifier to get a deterministic order.
+        """
+        return list(
+            sorted(notes, key=lambda x: (x.end_datetime, x.identifier), reverse=True)
+        )
 
     @staticmethod
     def _json_load(path: Path) -> List[dict]:
-        with path.open("r") as f:
+        with path.open("r", encoding="utf-8") as f:
             content = json.load(f, object_hook=_deserialize_datetime)
         return content
 
     @staticmethod
     def _json_dump(obj: List[dict], path: Path) -> None:
-        with path.open("w") as f:
-            json.dump(obj, f, cls=DatetimeJSONEncoder)
+        json_str = json.dumps(obj, cls=DatetimeJSONEncoder)
+        with path.open("w", encoding="utf-8") as f:
+            f.write(json_str)
 
     def __repr__(self) -> str:
         return f"Store('{self.path}')"
@@ -498,7 +604,7 @@ class Store(BaseStore):
 
 class DatetimeJSONEncoder(JSONEncoder):
     """Encodes datetime objects as a dictionary
-    with key "_isoformat" and the isoformat representation
+    with key "_datetime" and a string representation
     of the datetime as value.
 
     Idea for this encoder class comes from
@@ -507,7 +613,7 @@ class DatetimeJSONEncoder(JSONEncoder):
 
     def default(self, obj):
         if isinstance(obj, datetime):
-            return {"_isoformat": _format_datetime(obj)}
+            return {"_datetime": _format_datetime(obj)}
         return super().default(obj)
 
 
@@ -518,18 +624,18 @@ def _deserialize_datetime(obj):
     Idea for this function comes from
     https://stackoverflow.com/a/52838324
     """
-    _isoformat = obj.get("_isoformat")
-    if _isoformat is not None:
-        return _parse_datetime(_isoformat)
+    _datetime = obj.get("_datetime")
+    if _datetime is not None:
+        return _parse_datetime(_datetime)
     return obj
 
 
 def _format_datetime(dt: datetime) -> str:
-    return dt.isoformat()
+    return dt.strftime(DATETIME_STRING_FORMAT)
 
 
 def _parse_datetime(dt_str: str) -> datetime:
-    return datetime.strptime(dt_str, "%Y-%m-%dT%H:%M:%S.%f")
+    return datetime.strptime(dt_str, DATETIME_STRING_FORMAT)
 
 
 def _convert_to_path(path: Union[str, Path]) -> Path:
